"""
Classic cart-pole system implemented by Rich Sutton et al.
Copied from http://incompleteideas.net/sutton/book/code/pole.c
permalink: https://perma.cc/C9ZM-652R

Modified for swing up problem with constraints
"""

import math
import gym
from gym import spaces, logger
from gym.utils import seeding
import numpy as np

class CartSafeEnv(gym.Env):
    """
    Description:
        A pole is attached by an un-actuated joint to a cart, which moves along a frictionless track.
        The pendulum starts hanging down, and the goal is to swing it up by increasing and reducing the cart's velocity.

    Source:
        This environment is a modified version of the cart-pole problem described by Barto, Sutton, and Anderson

    Observation:
        Type: Box(4)
        Num	Observation                 Min         Max
        0	Cart Position             -4.8            4.8
        1	Cart Velocity             -Inf            Inf
        2	Pole Angle                 0              2*math.pi
        3	Pole Velocity At Tip      -Inf            Inf

        Note: angle state is only up to 2*Pi, meaning e.g. 1*Pi is the same state as 3*Pi

    Actions:
        Type: Discrete(2)
        Num	Action
        0	Push cart to the left
        1	Push cart to the right

        Note: The amount the velocity that is reduced or increased is not fixed; it depends on the angle the pole is pointing. This is because the center of gravity of the pole increases the amount of energy needed to move the cart underneath it

    Reward:
        Since we want the agent to learn to swing up our Reward is 1+cos(Pole Angle) for every step taken,
        including the termination step

    Constraint Cost:
        We constrain the position of the cart. Every time step that the cart violates these constraints we return an
        immediate cost of 1 for the violated constraint. For Constrained Markov Decision Problems (CMDPs) we typically
        want to have a threshold for the expected cumulative constraint costs.
        The immediate constraint costs are returned inside the info dict.
        Constraints:
            -1 < Cart Position < 1


    Starting State:
        All observations are assigned a uniform random value in [-0.05..0.05], for the angle we add pi such that the pendulum
        starts at the bottom and the environment is consequently more difficult

    Episode Termination:
        Cart Position is more than 2.4 (center of the cart reaches the edge of the display)
        Episode length is greater than 300
        Solved Requirements
        Considered solved when the average reward is greater than or equal to 520 over 100 consecutive trials.
            A reward of 520 corresponds approx to a lower bound for an epsiode of 300 steps with 75% of the time in between an angle of [-12,12] degrees.
    """

    metadata = {
        'render.modes': ['human', 'rgb_array'],
        'video.frames_per_second' : 50
    }

    def __init__(self):
        self.gravity = 9.8
        self.masscart = 1.0
        self.masspole = 0.1
        self.total_mass = (self.masspole + self.masscart)
        self.length = 0.5 # actually half the pole's length
        self.polemass_length = (self.masspole * self.length)
        self.force_mag = 10.0
        self.tau = 0.02  # seconds between state updates
        self.kinematics_integrator = 'euler'

        # Angle at which to fail the episode
        # self.theta_threshold_radians = 12 * 2 * math.pi / 360
        self.x_threshold = 2.4

        self.x_constraint = 1.0

        # Angle limit set to 2 * theta_threshold_radians so failing observation is still within bounds
        high = np.array([
            self.x_threshold * 2,
            np.finfo(np.float32).max,
            # self.theta_threshold_radians * 2,
            2*math.pi,
            np.finfo(np.float32).max])

        low = -np.array([
            self.x_threshold * 2,
            np.finfo(np.float32).max,
            # self.theta_threshold_radians * 2,
            0,
            np.finfo(np.float32).max])

        self.action_space = spaces.Discrete(2)
        self.observation_space = spaces.Box(low, high, dtype=np.float32)

        self.seed()
        self.viewer = None
        self.state = None

        self.steps_beyond_done = None
        self.episodic_cost = []

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def constraint_cost(self, x, x_dot):
        if np.abs(x) > self.x_constraint:
            x_cost = 1
        else:
            x_cost = 0

        # if np.abs(x_dot) > self.x_dot_constraint:
        #     x_dot_cost = 1
        # else:
        #     x_dot_cost = 0
        return [x_cost]


    def step(self, action):
        assert self.action_space.contains(action), "%r (%s) invalid"%(action, type(action))
        state = self.state
        x, x_dot, theta, theta_dot = state
        force = self.force_mag if action==1 else -self.force_mag
        costheta = math.cos(theta)
        sintheta = math.sin(theta)
        temp = (force + self.polemass_length * theta_dot * theta_dot * sintheta) / self.total_mass
        thetaacc = (self.gravity * sintheta - costheta* temp) / (self.length * (4.0/3.0 - self.masspole * costheta * costheta / self.total_mass))
        xacc  = temp - self.polemass_length * thetaacc * costheta / self.total_mass
        if self.kinematics_integrator == 'euler':
            x  = x + self.tau * x_dot
            x_dot = x_dot + self.tau * xacc
            theta = theta + self.tau * theta_dot
            theta_dot = theta_dot + self.tau * thetaacc
        else: # semi-implicit euler
            x_dot = x_dot + self.tau * xacc
            x  = x + self.tau * x_dot
            theta_dot = theta_dot + self.tau * thetaacc
            theta = theta + self.tau * theta_dot
        theta = theta % (2*math.pi)
        self.state = (x,x_dot,theta,theta_dot)
        done =  x < -self.x_threshold \
                or x > self.x_threshold
                # or theta < -self.theta_threshold_radians \
                # or theta > self.theta_threshold_radians
        done = bool(done)

        if not done:
            reward = 1+np.math.cos(theta)
        elif self.steps_beyond_done is None:
            # Pole just fell!
            self.steps_beyond_done = 0
            reward = 1+np.math.cos(theta)
        else:
            if self.steps_beyond_done == 0:
                logger.warn("You are calling 'step()' even though this environment has already returned done = True. You should always call 'reset()' once you receive 'done = True' -- any further steps are undefined behavior.")
            self.steps_beyond_done += 1
            reward = -1.0

        constraint_costs = self.constraint_cost(x, x_dot)
        self.episodic_cost.append(constraint_costs)

        if done:
            cost = [np.sum(self.episodic_cost)]
            self.episodic_cost.clear()
            return np.array(self.state), reward, done, {'cost': cost}
        else:
            return np.array(self.state), reward, done, {'cost': [0]}

    def reset(self):
        self.state = self.np_random.uniform(low=-0.05, high=0.05, size=(4,))
        self.state[2] += math.pi
        self.steps_beyond_done = None
        return np.array(self.state)

    def render(self, mode='human'):
        screen_width = 600
        screen_height = 400

        world_width = self.x_threshold*2
        scale = screen_width/world_width
        carty = 100 # TOP OF CART
        polewidth = 10.0
        polelen = scale * (2 * self.length)
        cartwidth = 50.0
        cartheight = 30.0

        if self.viewer is None:
            from gym.envs.classic_control import rendering
            self.viewer = rendering.Viewer(screen_width, screen_height)
            l,r,t,b = -cartwidth/2, cartwidth/2, cartheight/2, -cartheight/2
            axleoffset =cartheight/4.0
            cart = rendering.FilledPolygon([(l,b), (l,t), (r,t), (r,b)])
            self.carttrans = rendering.Transform()
            cart.add_attr(self.carttrans)
            self.viewer.add_geom(cart)
            l,r,t,b = -polewidth/2,polewidth/2,polelen-polewidth/2,-polewidth/2
            pole = rendering.FilledPolygon([(l,b), (l,t), (r,t), (r,b)])
            pole.set_color(.8,.6,.4)
            self.poletrans = rendering.Transform(translation=(0, axleoffset))
            pole.add_attr(self.poletrans)
            pole.add_attr(self.carttrans)
            self.viewer.add_geom(pole)
            self.axle = rendering.make_circle(polewidth/2)
            self.axle.add_attr(self.poletrans)
            self.axle.add_attr(self.carttrans)
            self.axle.set_color(.5,.5,.8)
            self.viewer.add_geom(self.axle)
            self.track = rendering.Line((0,carty), (screen_width,carty))
            self.track.set_color(0,0,0)
            self.viewer.add_geom(self.track)

            self.constr_line_l = rendering.Line((-self.x_constraint*scale + screen_width/2., carty-100),
                                              (-self.x_constraint*scale + screen_width/2., carty+1000))
            self.constr_line_l.set_color(1,0,0)
            self.viewer.add_geom(self.constr_line_l)
            self.constr_line_r = rendering.Line((self.x_constraint*scale + screen_width/2., carty-100),
                                              (self.x_constraint*scale + screen_width/2., carty+1000))
            self.constr_line_r.set_color(1,0,0)
            self.viewer.add_geom(self.constr_line_r)


            self._pole_geom = pole

        if self.state is None: return None

        # Edit the pole polygon vertex
        pole = self._pole_geom
        l,r,t,b = -polewidth/2,polewidth/2,polelen-polewidth/2,-polewidth/2
        pole.v = [(l,b), (l,t), (r,t), (r,b)]

        x = self.state
        cartx = x[0]*scale+screen_width/2.0 # MIDDLE OF CART
        self.carttrans.set_translation(cartx, carty)
        self.poletrans.set_rotation(-x[2])

        return self.viewer.render(return_rgb_array = mode=='rgb_array')

    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None
